!---------------------------------------------------------------------------------------
! mod_settings
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!> mod_settings
!!
!---------------------------------------------------------------------------------------


module mod_settings

  use mod_var

  implicit none
  private

  public :: files_t, config_t, read_file_settings, read_config_settings, read_region_settings, parse_string

  ! files_t contains information on paths and files

  type :: files_t

    character(len=max_path_len) :: path_prior     ! path to prior fluxes
    character(len=max_path_len) :: path_obs       ! path to observations
    character(len=max_path_len) :: path_output    ! path to output
    character(len=max_name_len) :: suffix         ! observation file name suffix
    character(len=max_path_len) :: path_flexpart  ! path to flexpart output
    character(len=max_name_len) :: file_log       ! log file name
    character(len=max_path_len) :: file_regions   ! region definitions file
    character(len=max_name_len) :: varname_regs   ! regions variable name
    character(len=max_name_len) :: lonname_regs   ! longitude variable name
    character(len=max_name_len) :: latname_regs   ! latitude variable name
    character(len=max_path_len) :: file_lsm       ! land-sea mask file
    character(len=max_name_len) :: varname_lsm    ! land-sea variable name
    character(len=max_name_len) :: lonname_lsm    ! longitude variable name
    character(len=max_name_len) :: latname_lsm    ! latitude variable name
    character(len=max_path_len) :: file_ice       ! snow-ice mask file
    character(len=max_name_len) :: varname_ice    ! snow-ice variable name
    character(len=max_name_len) :: lonname_ice    ! longitude variable name
    character(len=max_name_len) :: latname_ice    ! latitude variable name
    character(len=max_name_len) :: filename_flx   ! ghg flux file name 
    character(len=max_name_len) :: varname_flx    ! ghg flux variable name 
    character(len=max_name_len) :: lonname_flx    ! longitude variable name 
    character(len=max_name_len) :: latname_flx    ! latitude variable name 
    character(len=max_name_len) :: timename_flx   ! time variable name 
    character(len=max_name_len) :: filenest_flx   ! ghg flux nest file name 
    character(len=max_name_len) :: varnest_flx    ! ghg flux nest variable name 
    character(len=max_name_len) :: lonnest_flx    ! longitude nest variable name 
    character(len=max_name_len) :: latnest_flx    ! latitude nest variable name 
    character(len=max_name_len) :: timenest_flx   ! time nest variable name 
    character(len=max_name_len) :: filename_nee   ! nee flux file name 
    character(len=max_name_len) :: varname_nee    ! nee flux variable name 
    character(len=max_name_len) :: lonname_nee    ! longitude variable name 
    character(len=max_name_len) :: latname_nee    ! latitude variable name 
    character(len=max_name_len) :: timename_nee   ! time variable name
    character(len=max_name_len) :: filenest_nee   ! nee flux nest file name 
    character(len=max_name_len) :: varnest_nee    ! nee flux nest variable name 
    character(len=max_name_len) :: lonnest_nee    ! longitude nest variable name 
    character(len=max_name_len) :: latnest_nee    ! latitude nest variable name 
    character(len=max_name_len) :: timenest_nee   ! time nest variable name 
    character(len=max_path_len) :: file_recept    ! receptor list file

  end type files_t

  ! config_t contains information on configuration settings

  type :: config_t
 
    character(len=3)            :: spec            ! species ('co2' or 'ghg')
    integer                     :: datei           ! start date (yyyymmdd)
    integer                     :: datef           ! end date (yyyymmdd)
    logical                     :: nested          ! use nested flexpart output (true or false)
    real                        :: w_edge_lon      ! lon of western edge of inversion grid 
    real                        :: s_edge_lat      ! lat of southern edge of inversion grid 
    real                        :: e_edge_lon      ! lon of eastern edge of inversion grid 
    real                        :: n_edge_lat      ! lat of northern edge of inversion grid 
    real                        :: xres            ! longitudinal grid resolution
    real                        :: yres            ! latitudinal grid resolution
    integer                     :: nt_flx          ! number of flux time steps

  end type config_t

  contains

    ! --------------------------------------------------
    ! check if input is numeric
    ! --------------------------------------------------

    logical function is_numeric(string)

      character(len=200), intent(in) :: string
      real :: var
      integer :: ierr

      read(string,*,iostat=ierr) var
      if(ierr.eq.0) then
        is_numeric=.true.
      else
        is_numeric=.false.
      endif

    end function is_numeric

    ! --------------------------------------------------
    ! check if input is logical
    ! --------------------------------------------------

    logical function is_logical(string)

      character(len=200), intent(in) :: string
      logical :: var
      integer :: ierr

      read(string,*,iostat=ierr) var
      if(ierr.eq.0) then
        is_logical=.true.
      else
        is_logical=.false.
      endif

    end function is_logical

    ! --------------------------------------------------
    ! split string
    ! --------------------------------------------------

    subroutine split(str,delims,before,sep)

    ! Routine finds the first instance of a character from 'delims' in the
    ! the string 'str'. The characters before the found delimiter are
    ! output in 'before'. The characters after the found delimiter are
    ! output in 'str'. The optional output character 'sep' contains the 
    ! found delimiter. A delimiter in 'str' is treated like an ordinary 
    ! character if it is preceded by a backslash (\). If the backslash 
    ! character is desired in 'str', then precede it with another backslash.

    character(len=*) :: str, delims, before
    character,optional :: sep
    logical :: pres
    character :: ch,cha
    integer :: i, k, ibsl, ipos, iposa, lenstr

    pres=present(sep)
    str=adjustl(str)
    lenstr=len_trim(str)
    if(lenstr == 0) return        ! string str is empty
    k=0
    ibsl=0                        ! backslash initially inactive
    before=' '
    do i=1,lenstr
      ch=str(i:i)
      if(ibsl == 1) then          ! backslash active
        k=k+1
        before(k:k)=ch
        ibsl=0
        cycle
      end if
      if(ch == '\') then          ! backslash with backslash inactive
        k=k+1
        before(k:k)=ch
        ibsl=1
        cycle
      end if
      ipos=index(delims,ch)
      if(ipos == 0) then          ! character is not a delimiter
        k=k+1
        before(k:k)=ch
        cycle
      end if
      if(ch /= ' ') then          ! character is a delimiter that is not a space
        str=str(i+1:)
        if(pres) sep=ch
        exit
      end if
      cha=str(i+1:i+1)            ! character is a space delimiter
      iposa=index(delims,cha)
      if(iposa > 0) then          ! next character is a delimiter
        str=str(i+2:)
        if(pres) sep=cha
        exit
      else
        str=str(i+1:)
        if(pres) sep=ch
        exit
      end if
    end do
    if(i >= lenstr) str=''
    str=trim(str)
    before=trim(before)
    return

    end subroutine split

    ! --------------------------------------------------
    ! routine to identify and read in content
    ! --------------------------------------------------

    subroutine read_content (line, identifier, cc, cn, cl, match)

      character (len=200), intent (in) :: line, identifier
      character (len=200), intent (out) :: cc ! character content
      real(kind=8),        intent (out) :: cn ! numeric content
      logical,             intent (out) :: cl ! logical content
      logical,             intent (out) :: match

      integer :: n

      n = len_trim (identifier)

      ! default: line does not match, cc="", cn=-9900.0
      cc    = ""
      cn    = -9999.9
      match = .false.
      if ( len_trim (line) >= n ) then
        if ( line(:n) == identifier(:n) ) then
          cc = adjustl (line(n+1:))
          if ( is_numeric(cc) ) then
            read (line(n+1:),*) cn
          else
            if( is_logical(cc) ) read (line(n+1:),*) cl
          endif
          match = .true.
        end if
      end if

    end subroutine read_content

    ! --------------------------------------------------
    ! routine to parse a string
    ! --------------------------------------------------

    subroutine parse_string (str, delims, args, nargs)

    ! Parses the string 'str' into arguments args(1), ..., args(nargs) based on
    ! the delimiters contained in the string 'delims'. Preceding a delimiter in
    ! 'str' by a backslash (\) makes this particular instance not a delimiter.
    ! The integer output variable nargs contains the number of arguments found.

    character(len=*) :: str, delims
    character(len=len_trim(str)) :: strsav
    character(len=*), dimension(:) :: args
    integer :: i, na, lenstr, nargs

    strsav=str
    na=size(args)

    do i=1,na
      args(i)=' '
    end do
    nargs=0
    lenstr=len_trim(str)
    if(lenstr==0) return

    do
      if(len_trim(str) == 0) exit
      nargs=nargs+1
      call split(str,delims,args(nargs))
    end do
    str=strsav

    end subroutine parse_string

    ! --------------------------------------------------
    ! read file settings
    ! --------------------------------------------------

    subroutine read_file_settings(filename, files)

      character(len=200), intent(in)   :: filename
      type(files_t), intent(in out)    :: files

      integer            :: ierr, i, n
      character(len=200) :: line, identifier, cc
      real(kind=8)       :: cn
      logical            :: match, cl


     ! default values 
     files%file_ice = "" 

     ! open file
      open (100, file = trim (filename), status = 'old', iostat=ierr)
      if(ierr.gt.0) then
        write (*,*) 'ERROR: cannot find '//trim(filename)
        stop
      endif

      ! loop over lines
      read_loop: do
        read (100, fmt='(A)', iostat=ierr) line

        ! first check if line exists or eof was reached
        if ( ierr.gt.0 ) then
          write (*,*) 'ERROR: check '//trim(filename)
          exit read_loop
        else if ( ierr.lt.0 ) then
          exit read_loop
        else

          ! omit comments ("#") and empty lines
          if ( index (trim (line), "#") .eq. 1 ) cycle read_loop
          if ( len   (trim (line))      .eq. 0 ) cycle read_loop

          ! read path settings
          identifier = "path_prior:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%path_prior = cc
          identifier = "path_obs:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%path_obs = cc
          identifier = "suffix:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%suffix = cc
          identifier = "path_flexpart:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%path_flexpart = cc
          identifier = "path_output:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%path_output = cc
          identifier = "file_log:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%file_log = cc

          ! read region mask settings
          identifier = "file_regions:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%file_regions = cc
          identifier = "varname_regs:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%varname_regs = cc
          identifier = "lonname_regs:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%lonname_regs = cc
          identifier = "latname_regs:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%latname_regs = cc

          ! read land-sea mask settings
          identifier = "file_lsm:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%file_lsm = cc
          identifier = "varname_lsm:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%varname_lsm = cc
          identifier = "lonname_lsm:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%lonname_lsm = cc
          identifier = "latname_lsm:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%latname_lsm = cc

          ! read snow-ice mask settings
          identifier = "file_ice:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%file_ice = cc
          identifier = "varname_ice:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%varname_ice = cc
          identifier = "lonname_ice:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%lonname_ice = cc
          identifier = "latname_ice:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%latname_ice = cc

          ! read flux prior settings
          identifier = "filename_flx:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%filename_flx = cc
          identifier = "varname_flx:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%varname_flx = cc
          identifier = "lonname_flx:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%lonname_flx = cc
          identifier = "latname_flx:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%latname_flx = cc
          identifier = "timename_flx:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%timename_flx = cc
          identifier = "filename_nee:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%filename_nee = cc
          identifier = "varname_nee:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%varname_nee = cc
          identifier = "lonname_nee:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%lonname_nee = cc
          identifier = "latname_nee:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%latname_nee = cc
          identifier = "timename_nee:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%timename_nee = cc
          identifier = "filenest_flx:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%filenest_flx = cc
          identifier = "varnest_flx:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%varnest_flx = cc
          identifier = "lonnest_flx:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%lonnest_flx = cc
          identifier = "latnest_flx:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%latnest_flx = cc
          identifier = "timenest_flx:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%timenest_flx = cc
          identifier = "filenest_nee:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%filenest_nee = cc
          identifier = "varnest_nee:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%varnest_nee = cc
          identifier = "lonnest_nee:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%lonnest_nee = cc
          identifier = "latnest_nee:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%latnest_nee = cc
          identifier = "timenest_nee:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%timenest_nee = cc

          ! read receptor file settings
          identifier = "file_recept:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) files%file_recept = cc

        endif

      end do read_loop

    end subroutine read_file_settings

    ! --------------------------------------------------
    ! read config settings
    ! --------------------------------------------------

    subroutine read_config_settings(filename, config)

      character(len=200), intent(in)   :: filename
      type(config_t), intent(in out)   :: config

      integer            :: ierr, i, n
      character(len=200) :: line, identifier, cc
      real(kind=8)       :: cn
      logical            :: match, cl
      character(len=20), dimension(3) :: temp

     ! open file
      open (100, file = trim (filename), status = 'old', iostat=ierr)
      if(ierr.gt.0) then
        write(*,*) 'ERROR: cannot find '//trim(filename)
        stop
      endif

      ! loop over lines
      read_loop: do
        read (100, fmt='(A)', iostat=ierr) line

        ! first check if line exists or eof was reached
        if ( ierr.gt.0 ) then
          write (*,*) 'ERROR: check '//trim(filename)
          exit read_loop
        else if ( ierr.lt.0 ) then
          exit read_loop
        else

          ! omit comments ("#") and empty lines
          if ( index (trim (line), "#") .eq. 1 ) cycle read_loop
          if ( len   (trim (line))      .eq. 0 ) cycle read_loop

          ! read species
          identifier = "spec:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) config%spec = cc

          ! read date settings
          identifier = "datei:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) config%datei = int(cn)
          identifier = "datef:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) config%datef = int(cn)

          ! read inversion domain settings
          identifier = "nested:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) config%nested = cl
          identifier = "w_edge_lon:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) config%w_edge_lon = real(cn)
          identifier = "s_edge_lat:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) config%s_edge_lat = real(cn)
          identifier = "e_edge_lon:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) config%e_edge_lon = real(cn)
          identifier = "n_edge_lat:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) config%n_edge_lat = real(cn)
          identifier = "xres:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) config%xres = real(cn)
          identifier = "yres:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) config%yres = real(cn)
          identifier = "nt_flx:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) config%nt_flx = int(cn)

        endif

      end do read_loop

    end subroutine read_config_settings

    ! --------------------------------------------------
    ! read region settings
    ! --------------------------------------------------

    subroutine read_region_settings(filename)

      character(len=200), intent(in)   :: filename

      integer            :: ierr, i, n
      character(len=200) :: line, identifier, cc
      real(kind=8)       :: cn
      logical            :: match, cl
      character(len=20), dimension(:), allocatable :: temp

     ! open file
      open (100, file = trim (filename), status = 'old', iostat=ierr)
      if(ierr.gt.0) then
        write(*,*) 'ERROR: cannot find '//trim(filename)
        stop
      endif

      ! loop over lines
      read_loop: do
        read (100, fmt='(A)', iostat=ierr) line

        ! first check if line exists or eof was reached
        if ( ierr.gt.0 ) then
          write (*,*) 'ERROR: check '//trim(filename)
          exit read_loop
        else if ( ierr.lt.0 ) then
          exit read_loop
        else

          ! omit comments ("#") and empty lines
          if ( index (trim (line), "#") .eq. 1 ) cycle read_loop
          if ( len   (trim (line))      .eq. 0 ) cycle read_loop

          ! read region settings
          identifier = "useflux:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) useflux = cl
          identifier = "uselanduse:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) uselanduse = cl
          identifier = "cutoff:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) cutoff = cn
          identifier = "nmap:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) nmap = int(cn)
          identifier = "mapdec:"
          call read_content (line, identifier, cc, cn, cl, match)
          if ( match ) then
            if (.not.allocated(mapdec)) allocate(mapdec(nmap))
            allocate(temp(nmap))
            call parse_string (cc, ",", temp(:), n)
            do i = 1, nmap
              read(temp(i),*) mapdec(i)
            enddo
            deallocate( temp )
          endif

        endif

      end do read_loop

    end subroutine read_region_settings

    ! --------------------------------------------------

end module mod_settings

